/*=============================================================================*
 *  Copyright 2004 The Apache Software Foundation
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *=============================================================================*/
package org.apache.ws.resource.impl;

import commonj.timers.Timer;
import commonj.timers.TimerManager;
import org.apache.commons.collections.map.ReferenceMap;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.ws.Soap1_1Constants;
import org.apache.ws.addressing.EndpointReference;
import org.apache.ws.addressing.XmlBeansEndpointReference;
import org.apache.ws.resource.IllegalResourceTypeException;
import org.apache.ws.resource.JndiConstants;
import org.apache.ws.resource.PersistentResource;
import org.apache.ws.resource.PropertiesResource;
import org.apache.ws.resource.Resource;
import org.apache.ws.resource.ResourceContext;
import org.apache.ws.resource.ResourceCreationEvent;
import org.apache.ws.resource.ResourceCreationListener;
import org.apache.ws.resource.ResourceDestructionEvent;
import org.apache.ws.resource.ResourceDestructionListener;
import org.apache.ws.resource.ResourceException;
import org.apache.ws.resource.ResourceHome;
import org.apache.ws.resource.ResourceUnknownException;
import org.apache.ws.resource.faults.FaultException;
import org.apache.ws.resource.i18n.Keys;
import org.apache.ws.resource.i18n.MessagesImpl;
import org.apache.ws.resource.lifetime.ScheduledResourceTerminationResource;
import org.apache.ws.util.Cache;
import org.apache.ws.util.NameUtils;
import org.apache.ws.util.i18n.Messages;
import org.apache.ws.util.jndi.Initializable;
import org.apache.ws.util.jndi.JNDIUtils;
import org.apache.ws.util.lock.Lock;
import org.apache.ws.util.lock.LockManager;
import org.apache.ws.util.platform.JaxRpcPlatform;
import javax.naming.Context;
import javax.naming.InitialContext;
import javax.naming.NamingException;
import javax.xml.namespace.QName;
import javax.xml.rpc.JAXRPCException;
import javax.xml.soap.SOAPException;
import javax.xml.soap.SOAPHeader;
import javax.xml.soap.SOAPHeaderElement;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

/**
 * TODO: Update these Javadocs
 * <p/>
 * LOG-DONE An implementation of the <code>ResourceHome</code> interface. This implementation was designed to work with
 * resources that implement the {@link PersistentResource PersistenceCallback} interface as well as memory resident
 * resources. If the resource class implements the {@link PersistentResource PersistenceCallback} interface
 * <code>SoftReference</code>s will be used to recycle resource objects. The resource class implementation is
 * responsible for saving its state to disk. This implementation will <b>not</b> call {@link PersistentResource#store()
 * PersistenceCallback.store()}. The resource implementation must have a default constructor. <br><br> Configuration
 * options: <ul> <li> sweeperDelay - configures how often the resource sweeper runs in msec. By default the resource
 * sweeper runs every minute. For example:
 * <pre>
 *    &lt;parameter&gt;
 *     &lt;name&gt;sweeperDelay&lt;/name&gt;
 *     &lt;value&gt;60000&lt;/value&gt;
 *    &lt;/parameter&gt;
 * </pre>
 * <li> resourceClass - configures the name of the resource class. For example:
 * <pre>
 *    &lt;parameter&gt;
 *     &lt;name&gt;resourceClass&lt;/name&gt;
 *     &lt;value&gt;org.globus.wsrf.samples.counter.PersistentCounter&lt;/value&gt;
 *    &lt;/parameter&gt;
 * </pre>
 * <li> resourceKeyType - configures the key type class. By default <code>java.lang.String</code> is used. For example:
 * <pre>
 *    &lt;parameter&gt;
 *     &lt;name&gt;resourceKeyType&lt;/name&gt;
 *     &lt;value&gt;java.lang.Integer&lt;/value&gt;
 *    &lt;/parameter&gt;
 * </pre>
 * <li> resourceKeyName - configures the key name. For example:
 * <pre>
 *    &lt;parameter&gt;
 *     &lt;name&gt;resourceKeyName&lt;/name&gt;
 *     &lt;value&gt;{http://counter.com}CounterKey&lt;/value&gt;
 *    &lt;/parameter&gt;
 * </pre>
 * </ul> <br> <b>Note:</b> Must be deployed with <code>org.globus.wsrf.jndi.BeanFactory</code> in JNDI or user must
 * first call {@link #init() initialize()} method. Also when overriding the {@link #init() initialize()} method make
 * sure to call <code>super.initialize();</code>.
 *
 * @author Globus, Ian Springer
 */
public abstract class AbstractResourceHome
   implements ResourceHome,
              Initializable
{
   private static final int DEFAULT_SWEEPER_DELAY = 60000;
   private static final Log LOG = LogFactory.getLog( AbstractResourceHome.class );

   /** DOCUMENT_ME */
   public static final Messages MSG = MessagesImpl.getInstance(  );

   /**
    * A static key used only for registering Singleton Resources
    */
   private static final Object SINGLETON_KEY         = new Object(  );
   private static final String PLACEHOLDER_IPADDRESS = "$IP_ADDRESS$";
   private static final String PLACEHOLDER_HOSTNAME  = "$HOST_NAME$";

   /**
    * DOCUMENT_ME
    */
   protected Map m_resources;

   /**
    * DOCUMENT_ME
    */
   protected String m_resourceClassName;

   /** DOCUMENT_ME */
   protected boolean m_resourceIsPersistent;
   private QName     m_resourceIdRefParamName;
   private String    m_wsdlTargetNamespace;
   private String    m_serviceClassName;

   /**
    * The baseURL obtained via JNDI config, used as the basis for the endpoint URL.
    */
   private String m_baseWebappUrl;

   /**
    * DOCUMENT_ME
    */
   protected LockManager m_lockManager;
   private String        m_cacheLocation;
   private Cache         m_cache;
   private long          m_sweeperDelay         = DEFAULT_SWEEPER_DELAY;
   private Sweeper       m_sweeper;
   private boolean       m_initialized;
   private Class         m_resourceClass;
   private List          m_creationListeners    = new ArrayList(  );
   private List          m_destructionListeners = new ArrayList(  );

   /**
    * DOCUMENT_ME
    *
    * @param jndiLocation DOCUMENT_ME
    */
   public void setCacheLocation( String jndiLocation )
   {
      m_cacheLocation = jndiLocation;
   }

   /**
    * DOCUMENT_ME
    *
    * @return DOCUMENT_ME
    */
   public String getCacheLocation(  )
   {
      return m_cacheLocation;
   }

   /**
    * DOCUMENT_ME
    *
    * @return DOCUMENT_ME
    */
   public abstract org.apache.ws.resource.properties.NamespaceVersionHolder getNamespaceVersionHolder(  );

   /**
    * Returns the EndpointReference associated with this Resource. Only the required fields will be filled in (i.e.
    * Address) AND the ReferenceProperty for the ResourceKey (if not a singleton)
    * <p/>
    * If the resourceKey is not equal to null ( not a singleton ), the reference properties will contain the key.
    *
    * @param resourceId the resource identifier, or null if the resource is a singleton
    * @return an EndpointReference (EPR) for the resource
    */
   public EndpointReference getEndpointReference( Object resourceId )
   {
      String endpointAddress =
         JaxRpcPlatform.getJaxRpcPlatform(  ).getEndpointUrl( getBaseWebappUrl(  ),
                                                              getServiceName(  ).getLocalPart(  ) );
      String wsAddressingURI = getNamespaceVersionHolder(  ).getAddressingNamespace(  );

      XmlBeansEndpointReference xmlBeansEndpointReference =
         new XmlBeansEndpointReference( endpointAddress, wsAddressingURI );
      xmlBeansEndpointReference.setResourceIdentifierReferenceParameterName( getResourceIdentifierReferenceParameterQName(  ) );
      xmlBeansEndpointReference.setResourceIdentifier( resourceId );
      xmlBeansEndpointReference.setPortTypeQName( getPortType(  ) );
      xmlBeansEndpointReference.setServicePortName( getServicePortName(  ) );
      xmlBeansEndpointReference.setServiceQName( getServiceName(  ) );
      return xmlBeansEndpointReference;
   }

   /**
    * DOCUMENT_ME
    *
    * @return DOCUMENT_ME
    */
   public abstract QName getPortType(  );

   /**
    * DOCUMENT_ME
    *
    * @param resourceClass DOCUMENT_ME
    */
   public void setResourceClassName( String resourceClass )
   {
      m_resourceClassName = resourceClass;
   }

   /**
    * DOCUMENT_ME
    *
    * @return DOCUMENT_ME
    */
   public String getResourceClassName(  )
   {
      return m_resourceClassName;
   }

   /**
    * DOCUMENT_ME
    *
    * @param className DOCUMENT_ME
    */
   public void setServiceClassName( String className )
   {
      m_serviceClassName = className;
   }

   /**
    * DOCUMENT_ME
    *
    * @return DOCUMENT_ME
    */
   public String getServiceClassName(  )
   {
      return m_serviceClassName;
   }

   /**
    * DOCUMENT_ME
    *
    * @return DOCUMENT_ME
    */
   public abstract QName getServiceName(  );

   /**
    * DOCUMENT_ME
    *
    * @return DOCUMENT_ME
    */
   public abstract String getServicePortName(  );

   /**
    * Sets the base webapp url.
    * <p/>
    * The URL should contin the webapp context name. (i.e. http://127.0.0.1:8080/wsrf)
    * <p/>
    * The URL may contain one of the following:
    * <p/>
    * <ol>
    * <li>the marker &lt;IP_ADDRESS> in which case we will attempt to determine the IP address at runtime.
    * i.e. http://&lt;IP_ADDRESS>:8080/wsrf (Do not use on multi-homed systems)</li>
    * <li>the marker &lt;HOST_NAME> in which case we will attempt to determine the host name at runtime
    * i.e. http://&lt;HOST_NAME>:8080/wsrf   </li>
    * <li>NO Marker in which case whatever value you add will be static and used.
    * i.e. http://myhostname:8080/wsrf  </li>
    * </ol>
    *
    * @param baseWebappUrl The base webapp url containing the webapp context. (i.e. http://127.0.0.1:8080/wsrf)
    */
   public void setBaseWebappUrl( String baseWebappUrl )
   { //todo we may need to figure out a way to stop everything if this method fails.
     //the url is nonexistant
      if ( ( baseWebappUrl == null ) || baseWebappUrl.equals( "" ) )
      {
         try
         {
            m_baseWebappUrl = getDefaultUrl(  );
            LOG.fatal( "The baseWebappUrl from the jndi-config.xml file was invalid! Defaulting to: "
                       + m_baseWebappUrl );
         }
         catch ( java.net.UnknownHostException e )
         {
            m_baseWebappUrl = "http://127.0.0.1:8080/wsrf";
            LOG.fatal( "The baseWebappUrl from the jndi-config.xml file was invalid! Unable to determine host IP address using java.net.InetAddress.getLocalHost().getHostAddress() defaulting to: "
                       + m_baseWebappUrl, e );
            return;
         }
      }

      //the url is not to be modified
      else if ( ( baseWebappUrl.indexOf( PLACEHOLDER_IPADDRESS ) == -1 )
                && ( baseWebappUrl.indexOf( PLACEHOLDER_HOSTNAME ) == -1 ) )
      {
         m_baseWebappUrl = baseWebappUrl;
         return;
      }

      //url will need to be built
      //define localhost to be used
      java.net.InetAddress localHost = null;

      try
      {
         localHost = java.net.InetAddress.getLocalHost(  );
      }
      catch ( java.net.UnknownHostException e )
      {
         LOG.fatal( "Unable to get the InetAddress for localhost using InetAddress.getLocalHost().  Defaults will be used.  This may cause problems with EndpointReferences in Resources." );
      }

      if ( baseWebappUrl.indexOf( PLACEHOLDER_IPADDRESS ) > -1 )
      {
         String ipaddress = null;
         if ( localHost == null )
         {
            ipaddress = "127.0.0.1";
         }
         else
         {
            ipaddress = localHost.getHostAddress(  );
         }

         m_baseWebappUrl =
            baseWebappUrl.substring( 0,
                                     baseWebappUrl.indexOf( PLACEHOLDER_IPADDRESS ) ) + ipaddress
            + baseWebappUrl.substring( baseWebappUrl.indexOf( PLACEHOLDER_IPADDRESS )
                                       + PLACEHOLDER_IPADDRESS.length(  ) );
         return;
      } //end if

      else if ( baseWebappUrl.indexOf( PLACEHOLDER_HOSTNAME ) > -1 )
      {
         String hostname = null;
         if ( localHost == null )
         {
            hostname = "127.0.0.1";
         }
         else
         {
            hostname = localHost.getHostName(  );
         }

         m_baseWebappUrl =
            baseWebappUrl.substring( 0,
                                     baseWebappUrl.indexOf( PLACEHOLDER_HOSTNAME ) ) + hostname
            + baseWebappUrl.substring( baseWebappUrl.indexOf( PLACEHOLDER_HOSTNAME )
                                       + PLACEHOLDER_HOSTNAME.length(  ) );
         return;
      } //end else if
   }

   /**
    * Returns the base webapp url which includes the webapp context. (i.e. http://127.0.0.1:8080/wsrf)
    *
    * @return base webapp url
    */
   public String getBaseWebappUrl(  )
   {
      return m_baseWebappUrl;
   }

   /**
    * Returns true if the home has been initialized, else false
    *
    * @return true if the home has been initialized, else false
    */
   public boolean isInitialized(  )
   {
      return m_initialized;
   }

   /**
    * DOCUMENT_ME
    *
    * @param name DOCUMENT_ME
    */
   public void setResourceIdentifierReferenceParameterName( String name )
   {
      m_resourceIdRefParamName = QName.valueOf( name );
   }

   /**
    * DOCUMENT_ME
    *
    * @return DOCUMENT_ME
    */
   public String getResourceIdentifierReferenceParameterName(  )
   {
      //case of singleton...
      if ( m_resourceIdRefParamName == null )
      {
         return null;
      }

      return m_resourceIdRefParamName.toString(  );
   }

   /**
    * DOCUMENT_ME
    *
    * @return DOCUMENT_ME
    */
   public QName getResourceIdentifierReferenceParameterQName(  )
   {
      return m_resourceIdRefParamName;
   }

   /**
    * DOCUMENT_ME
    *
    * @param delay DOCUMENT_ME
    */
   public void setSweeperDelay( long delay )
   {
      m_sweeperDelay = delay;
   }

   /**
    * DOCUMENT_ME
    *
    * @return DOCUMENT_ME
    */
   public long getSweeperDelay(  )
   {
      return m_sweeperDelay;
   }

   /**
    * DOCUMENT_ME
    *
    * @param targetNamespace DOCUMENT_ME
    */
   public void setWsdlTargetNamespace( String targetNamespace )
   {
      m_wsdlTargetNamespace = targetNamespace;
   }

   /**
    * DOCUMENT_ME
    *
    * @return DOCUMENT_ME
    */
   public String getWsdlTargetNamespace(  )
   {
      return m_wsdlTargetNamespace;
   }

   /**
    * @see ResourceHome#add(org.apache.ws.resource.Resource)
    */
   public boolean add( Resource resource )
   throws IllegalResourceTypeException
   {
      boolean resourceWasAdded = false;
      if ( !m_resourceClass.isAssignableFrom( resource.getClass(  ) ) )
      {
         throw new IllegalResourceTypeException( "This home only supports resources of type "
                                                 + m_resourceClassName );
      }

      synchronized ( m_resources )
      {
         Object key = getNonNullKey( resource.getID(  ) );
         if ( resource.getEndpointReference(  ) == null )
         {
            resource.setEndpointReference( getEndpointReference( key ) );
         }

         if ( !m_resources.containsKey( key ) )
         {
            LOG.debug( MSG.getMessage( Keys.ADDING_RESOURCE_WITH_ID,
                                       resource.getID(  ),
                                       getClass(  ).getName(  ) ) );
            m_resources.put( key, resource );
            scheduleSweeperTask(  );
            notifyResourceCreatedListeners( resource );
            updateCache( resource );
            resourceWasAdded = true;
         }
      }

      return resourceWasAdded;
   }

   /**
    * Adds a listener for ResourceCreationEvents
    *
    * @param listener
    */
   public void addResourceCreationListener( ResourceCreationListener listener )
   {
      m_creationListeners.add( listener );
   }

   /**
    * Adds a listener for ResourceDestructionEvents
    *
    * @param listener
    */
   public void addResourceDestructionListener( ResourceDestructionListener listener )
   {
      m_destructionListeners.add( listener );
   }

   /**
    * DOCUMENT_ME
    *
    * @param context DOCUMENT_ME
    *
    * @return DOCUMENT_ME
    */
   public Object extractResourceIdentifier( ResourceContext context )
   {
      if ( getResourceIdentifierReferenceParameterName(  ) == null )
      {
         return null;
      }

      try
      {
         SOAPHeader soapHeader          = context.getSOAPMessage(  ).getSOAPHeader(  );
         Iterator   resourceIdElemsIter =
            soapHeader.getChildElements( NameUtils.toName( getResourceIdentifierReferenceParameterQName(  ) ) );
         if ( !resourceIdElemsIter.hasNext(  ) )
         {
            throw new FaultException( Soap1_1Constants.FAULT_CLIENT,
                                      "The expected resource identifier reference parameter named "
                                      + getResourceIdentifierReferenceParameterName(  )
                                      + " was not found in the SOAP header." );
         }

         SOAPHeaderElement resourceIdElem = (SOAPHeaderElement) resourceIdElemsIter.next(  );
         if ( resourceIdElemsIter.hasNext(  ) )
         {
            throw new FaultException( Soap1_1Constants.FAULT_CLIENT,
                                      "More than one resource identifier reference parameter named "
                                      + getResourceIdentifierReferenceParameterName(  )
                                      + " was found in the SOAP header. Exactly one is expected." );
         }

         return resourceIdElem.getValue(  );
      }
      catch ( SOAPException soape )
      {
         throw new JAXRPCException( soape );
      }
   }

   /**
    * @see ResourceHome#find(Object)
    */
   public Resource find( Object resourceId )
   throws ResourceUnknownException, 
          ResourceException
   {
      LOG.debug( MSG.getMessage( Keys.FINDING_RESOURCE_WITH_KEY,
                                 String.valueOf( resourceId ) ) );
      Resource resource = null;
      synchronized ( m_resources )
      {
         //Lock lock = acquireLock( resourceId );
         //try
         //{
         resource = get( resourceId );
         updateCache( resource );

         //}
         //finally
         //{
         //    lock.release();
         //}
      }

      return resource;
   }

   /**
    * Initializes this home. Should be called <em>after</em> setters have been called on all bean properties.
    *
    * @throws Exception on error
    */
   public void init(  )
   throws Exception
   {
      synchronized ( this )
      {
         if ( m_initialized )
         {
            return;
         }

         LOG.debug( MSG.getMessage( Keys.INIT_HOME ) );
         Class resourceClass = getResourceClass(  );
         if ( PersistentResource.class.isAssignableFrom( resourceClass ) )
         {
            m_resourceIsPersistent = true;
         }

         Context initialContext = new InitialContext(  );
         m_resources      = getResourceMap(  );
         m_lockManager    = new LockManager(  );
         if ( ScheduledResourceTerminationResource.class.isAssignableFrom( getResourceClass(  ) ) )
         {
            initSweeper( initialContext );
         }

         initCachePolicy( new InitialContext(  ) );
         m_initialized = true;
      }
   }

   /**
    * @see ResourceHome#remove(Object)
    */
   public void remove( Object resourceId )
   throws ResourceUnknownException, 
          ResourceException
   {
      Resource resource = null;
      Lock     lock = acquireLock( resourceId );
      try
      {
         resource = get( resourceId );

         try
         {
            resource.destroy(  );
         }
         catch ( RuntimeException re )
         {
            throw new ResourceException( MSG.getMessage( Keys.FAILED_TO_DESTROY_RESOURCE, resource, re ) );
         }

         Resource removedResource = (Resource) m_resources.remove( getNonNullKey( resourceId ) );
         if ( removedResource instanceof PropertiesResource )
         {
            notifyResourceDeletedListeners( removedResource );
         }

         LOG.debug( MSG.getMessage( Keys.REMOVED_RESOURCE_WITH_KEY,
                                    resource.getClass(  ).getName(  ),
                                    String.valueOf( resourceId ) ) );
         if ( m_cache != null )
         {
            m_cache.remove( resource );
         }
      }
      finally
      {
         lock.release(  );
      }
   }

   /**
    * Removes a listener for ResourceCreationEvents
    *
    * @param listener
    * @return true if the listener was removed, else false
    */
   public boolean removeResourceCreationListener( ResourceCreationListener listener )
   {
      return m_creationListeners.remove( listener );
   }

   /**
    * Removes a listener for ResourceDestructionEvents
    *
    * @param listener
    * @return true if the listener was removed, else false
    */
   public boolean removeResourceDestructionListener( ResourceDestructionListener listener )
   {
      return m_destructionListeners.remove( listener );
   }

   /**
    * DOCUMENT_ME
    *
    * @return DOCUMENT_ME
    *
    * @throws NamingException DOCUMENT_ME
    */
   protected abstract Map getResourceMap(  )
   throws NamingException;

   /**
    * DOCUMENT_ME
    *
    * @return DOCUMENT_ME
    *
    * @throws java.net.UnknownHostException DOCUMENT_ME
    */
   protected String getDefaultUrl(  )
   throws java.net.UnknownHostException
   {
      return "http://" + java.net.InetAddress.getLocalHost(  ).getHostAddress(  ) + ":8080/wsrf";
   }

   /**
    * This method uses reflection to create an instance of a Resource which contains an empty constructor.  It will not
    * work with Resources which do not have an empty constructor.
    * <p/>
    * Note: the returned Resource will an EndpointReference associated with it.
    *
    * @param id the resource identifier
    * @return a resource of the type associated with this home
    * @throws ResourceException
    * @throws IllegalStateException
    */
   protected Resource createInstance( Object id )
   throws ResourceException
   {
      LOG.debug( MSG.getMessage( Keys.CREATING_INSTANCE_WITH_KEY,
                                 String.valueOf( id ) ) );
      Resource resource;
      try
      {
         resource = (Resource) getResourceClass(  ).newInstance(  );
      }
      catch ( Exception e )
      {
         throw new ResourceException( e );
      }

      resource.setID( id );
      try
      {
         LOG.debug( MSG.getMessage( Keys.INIT_RESOURCE_LIFECYCLE_INSTANCE,
                                    resource.getClass(  ).getName(  ) ) );
         if ( resource.getEndpointReference(  ) == null ) //also checked in add(..) in case they sidestep this
         {
            resource.setEndpointReference( getEndpointReference( id ) );
         }

         resource.init(  );
      }
      catch ( RuntimeException re )
      {
         throw new ResourceException( MSG.getMessage( Keys.FAILED_TO_INIT_RESOURCE, resource, re ), re );
      }

      return resource;
   }

   /**
    * DOCUMENT_ME
    *
    * @param id DOCUMENT_ME
    * @return DOCUMENT_ME
    * @throws ResourceException DOCUMENT_ME
    */
   protected Resource createNewInstanceAndLoad( Object id )
   throws ResourceException
   {
      Resource resource = createInstance( id );
      LOG.debug( MSG.getMessage( Keys.LOADING_RESOURCE_FROM_PERSISTENCE,
                                 String.valueOf( id ) ) );
      ( (PersistentResource) resource ).load(  );
      if ( ResourceSweeper.isExpired( resource ) )
      {
         throw new ResourceUnknownException( getNonNullKey( id ),
                                             getServicePortName(  ) );
      }

      return resource;
   }

   /**
    * Creates a synschronized map for storing {@link Resource}s. If the resource is persistent, a {@link ReferenceMap}
    * is used, otherwise a {@link HashMap} is used. Provided for the convenience of subclasses, which are responsible
    * for creating a map of resources and maintaining a static reference to that map.
    *
    * @param isPersistent if the resource is persistent
    * @return a synschronized map for storing {@link Resource}s
    */
   protected static Map createResourceMap( boolean isPersistent )
   {
      Map resourceMap;
      if ( isPersistent )
      {
         resourceMap = new ReferenceMap( ReferenceMap.HARD, ReferenceMap.SOFT, true );
      }
      else
      {
         resourceMap = new HashMap(  );
      }

      return Collections.synchronizedMap( resourceMap );
   }

   /**
    * DOCUMENT_ME
    *
    * @param initialContext DOCUMENT_ME
    *
    * @throws NamingException DOCUMENT_ME
    */
   protected void initCachePolicy( Context initialContext )
   throws NamingException
   {
      if ( m_cacheLocation != null )
      {
         m_cache = (Cache) JNDIUtils.lookup( initialContext, m_cacheLocation, Cache.class );
      }
   }

   /**
    * DOCUMENT_ME
    *
    * @param obj DOCUMENT_ME
    *
    * @return DOCUMENT_ME
    */
   Object getNonNullKey( Object obj )
   {
      return ( obj != null ) ? obj : SINGLETON_KEY;
   }

   private Class getResourceClass(  )
   throws ResourceException
   {
      if ( m_resourceClass == null )
      {
         if ( m_resourceClassName == null )
         {
            throw new IllegalStateException( MSG.getMessage( Keys.RESOURCE_CLASSNAME_NULL ) );
         }

         try
         {
            m_resourceClass = Class.forName( m_resourceClassName );
         }
         catch ( ClassNotFoundException cnfe )
         {
            throw new ResourceException( MSG.getMessage( Keys.RESOURCE_CLASS_NOT_FOUND, m_resourceClassName ) );
         }

         if ( !Resource.class.isAssignableFrom( m_resourceClass ) )
         {
            throw new ResourceException( "Specified resource class '" + m_resourceClassName
                                         + "' does not implement the " + Resource.class.getName(  )
                                         + " interface." );
         }
      }

      return m_resourceClass;
   }

   private Lock acquireLock( Object resourceId )
   throws ResourceException
   {
      Lock lock = m_lockManager.getLock( getNonNullKey( resourceId ) );
      try
      {
         lock.acquire(  );
      }
      catch ( InterruptedException ie )
      {
         throw new ResourceException( ie );
      }

      return lock;
   }

   private Resource get( Object resourceId )
   throws ResourceException
   {
      LOG.debug( MSG.getMessage( Keys.GETTING_RESOURCE_WITH_ID,
                                 resourceId,
                                 getClass(  ).getName(  ) ) );
      Resource resource = (Resource) m_resources.get( getNonNullKey( resourceId ) );
      if ( resource == null )
      {
         if ( !m_resourceIsPersistent )
         {
            throw new ResourceUnknownException( resourceId,
                                                getServicePortName(  ) );
         }

         add( createNewInstanceAndLoad( resourceId ) );
      }

      return resource;
   }

   private void initSweeper( Context initialContext )
   throws NamingException
   {
      LOG.debug( MSG.getMessage( Keys.TIMER_LOOKUP_WITH_JNDI_NAME, JndiConstants.KEY_NAME_DEFAULT_TIMER ) );
      TimerManager timerManager = (TimerManager) initialContext.lookup( JndiConstants.KEY_NAME_DEFAULT_TIMER );

      // TimerManager timerManager = new TimerManagerImpl();
      m_sweeper = new Sweeper( this, m_resources, timerManager, m_sweeperDelay );
   }

   /**
    * This method is used to notify listeners a resource has been created.
    *
    * @param resource The Resource which was created
    */
   private void notifyResourceCreatedListeners( Resource resource )
   {
      for ( int i = 0; i < m_creationListeners.size(  ); i++ )
      {
         ResourceCreationListener resourceCreationListener =
            (ResourceCreationListener) m_creationListeners.get( i );
         resourceCreationListener.creationOccurred( new ResourceCreationEvent( resource ) );
      }
   }

   /**
    * This method is used to notify listeners a resource has been deleted.
    *
    * @param resource The EndpointReference for the Resource which was created
    */
   private void notifyResourceDeletedListeners( Resource resource )
   {
      for ( int i = 0; i < m_destructionListeners.size(  ); i++ )
      {
         ResourceDestructionListener resourceDestructionListener =
            (ResourceDestructionListener) m_destructionListeners.get( i );
         ResourceDestructionEvent    event = new ResourceDestructionEvent( resource );
         resourceDestructionListener.destructionOccurred( event );
      }
   }

   private void scheduleSweeperTask(  )
   {
      if ( m_sweeper != null )
      {
         m_sweeper.schedule(  );
      }
   }

   private void updateCache( Resource resource )
   {
      if ( m_cache != null )
      {
         m_cache.update( resource );
      }
   }

   /**
    * This ResourceSweeper implementation just returns the resources currently stored in the map. The reason is that
    * the sweeper doesn't have to reactivate/reload a persistent resource if the resource object was reclaimed. So
    * lifetime checks are not done on reclained resources. Lifetime checks have to be done on resource load.
    */
   private static class Sweeper
      extends ResourceSweeper
   {
      private TimerManager m_timerManager;
      private Timer        m_timer;
      private long         m_delay;

      /**
       * Creates a new {@link Sweeper} object.
       *
       * @param home         DOCUMENT_ME
       * @param resources    DOCUMENT_ME
       * @param timerManager DOCUMENT_ME
       * @param delay        DOCUMENT_ME
       */
      public Sweeper( ResourceHome home,
                      Map          resources,
                      TimerManager timerManager,
                      long         delay )
      {
         super( home, resources );
         m_timerManager    = timerManager;
         m_delay           = delay;
      }

      /**
       * DOCUMENT_ME
       *
       * @param timer DOCUMENT_ME
       */
      public void timerExpired( Timer timer )
      {
         super.timerExpired( timer );
         cancel(  );
         if ( !m_resources.isEmpty(  ) )
         {
            schedule(  );
         }
      }

      /**
       * Schedules this resource sweeper.
       */
      synchronized void schedule(  )
      {
         if ( m_timer == null )
         {
            LOG.debug( MSG.getMessage( Keys.SCHEDULE_RESOURCE_SWEEPER ) );
            m_timer = m_timerManager.schedule( this, m_delay );
         }
      }

      /**
       * Cancels this resource sweeper.
       */
      private synchronized void cancel(  )
      {
         if ( m_timer != null )
         {
            LOG.debug( MSG.getMessage( Keys.CANCEL_RESOURCE_SWEEPER ) );
            m_timer = null;
         }
      }
   }
}