/*=============================================================================*
 *  Copyright 2004 The Apache Software Foundation
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *=============================================================================*/
package org.apache.ws.util.platform;

import org.apache.ws.util.platform.axis.AxisJaxRpcPlatform;
import org.apache.ws.util.platform.weblogic.WeblogicJaxRpcPlatform;
import javax.xml.soap.SOAPException;
import javax.xml.soap.SOAPFactory;

/**
 * An Enumeration of JAX-RPC platform types supported by Apache WSRF.
 * Currently the supported platforms are Apache Axis ({@link #AXIS})
 * and BEA WebLogic Server ({@link #WEBLOGIC}).
 *
 * @author Ian Springer, Sal Campana
 */
public abstract class JaxRpcPlatform
{
   /**
    * Axis JaxRpcPlatform instance.
    */
   public static JaxRpcPlatform AXIS = new AxisJaxRpcPlatform(  );

   /**
    * Weblogic JaxRpcPlatform instance.
    */
   public static JaxRpcPlatform WEBLOGIC = new WeblogicJaxRpcPlatform(  );

   /**
    * Represents the platform detected in this classloader.
    */
   private static JaxRpcPlatform s_platformType;

   /**
    * Returns a short description of the platform.
    *
    * @return a short description of the platform
    */
   public abstract String getDescription(  );

   /**
    * Returns the platform-specific endpoint url for a service on a given platform.
    *
    * An example of this would be:</br>
    * baseWebappUrl = http://localhost:8080/wsrf</br>
    * serviceName = filesystem</br>
    *
    * On the Axis platform the endpoint URL is:  http://localhost:8080/wsrf/services/filesystem
    *
    * @param baseWebappUrl The url containing the webapp context (i.e. http://localhost:8080/wsrf)
    * @param serviceName The service name which is registered with the platform
    * @return The endpoint url for the service.
    */
   public abstract String getEndpointUrl( String baseWebappUrl,
                                          String serviceName );

   /**
    * Returns the SOAPFactoryImpl class name.
    *
    * @return SOAPFactoryImpl class name.
    */
   public abstract String getSoapFactoryImpl(  );

   /**
    * Returns the description of the JAX-RPC platform.
    *
    * @return String
    */
   public abstract String toString(  );

   /**
    * Returns the JaxRpcPlatform for this environment. This method is meant to be called from within the classloader
    * hosting the JAX-RPC engine.
    *
    * @return the JaxRpcPlatform for this classloader
    */
   public static synchronized JaxRpcPlatform getJaxRpcPlatform(  )
   {
      if ( s_platformType != null )
      {
         return s_platformType;
      }

      SOAPFactory soapFactory;
      try
      {
         soapFactory = SOAPFactory.newInstance(  );
      }
      catch ( SOAPException soape )
      {
         throw new RuntimeException( "JAX-RPC platform not found!" );
      }

      if ( soapFactory.getClass(  ).getName(  ).equals( AXIS.getSoapFactoryImpl(  ) ) )
      {
         s_platformType = AXIS;
      }
      else if ( soapFactory.getClass(  ).getName(  ).equals( WEBLOGIC.getSoapFactoryImpl(  ) ) )
      {
         s_platformType = WEBLOGIC;
      }

      if ( s_platformType == null )
      {
         throw new RuntimeException( "Unsupported JAX-RPC platform." );
      }

      return s_platformType;
   }
}